import collections
import datetime
import os

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db.models import JSONField
from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.db.models import Model, Q
from django.utils import timezone
from django.utils.html import strip_tags
from django.utils.module_loading import import_string
from django.template.loader import render_to_string

from django_q.models import Schedule
from django_q.tasks import async_task, schedule

from .util import PGAdvisoryLock


class NotificationTypeChoices(models.TextChoices):
    EMAIL = 'email'
    SMS = 'sms'


class Subscription(models.Model):
    sender_type = models.ForeignKey(
        ContentType,
        null=True,
        blank=True,
        related_name="notification_subscription_sender_types",
        on_delete=models.CASCADE
    )
    sender_id = models.PositiveIntegerField(null=True)
    sender = GenericForeignKey('sender_type', 'sender_id')

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    notification_name = models.CharField(max_length=100)
    is_active = models.BooleanField(default=True)
    digest = models.BooleanField(default=False)

    class Meta:
        app_label = 'notificationcenter'
        unique_together = ('user', 'notification_name', 'sender_type', 'sender_id', 'digest',)


class Notification(models.Model):
    """
    A Notification object contains the data necessary to send an email notification to users
    who are subscribed to receive notifications of type 'name' and, optionally, that relate
    to an associated 'sender' object.

    The 'context' property is intended to contain info about the nature of the notification.
    A common usage would be to include the ID of the user who triggered the action
    responsible for creating the notification.
    """
    sender_type = models.ForeignKey(
        ContentType,
        null=True,
        blank=True,
        related_name="notification_sender_types",
        on_delete=models.CASCADE
    )
    sender_id = models.PositiveIntegerField(null=True)
    sender = GenericForeignKey('sender_type', 'sender_id')

    context = JSONField(default=dict)

    name = models.CharField(max_length=100)
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    date_created = models.DateTimeField(auto_now_add=True)
    date_sent = models.DateTimeField(blank=True, null=True)
    date_digested = models.DateTimeField(blank=True, null=True)

    def mark_as_sent(self, recipients=None):
        self.date_sent = timezone.now()
        self.save()

#-------------------------------------------------------------------------------
class NotificationCenter(object):
    def __init__(self):
        self.from_email = settings.NOTIFICATION_CENTER_SENDER_EMAIL

        self.implicit_subscription_rules = []
        self.alter_recipient_rules = []

        self.digest_implicit_subscription_rules = []
        self.digest_alter_recipient_rules = []

        # The default email address used for a subscribed user is user.email
        # An email_accessor function may be assigned to customize this behavior.
        self.email_accessor = lambda user: user.email

        # By default there is no accessor for phone numbers; must be supplied via settings.
        self.phone_accessor = lambda user: None

    #--------------------------------------------------------------------------
    def add_implicit_subscription_rule(self, notification_name, rule_func, digest=False):
        if not digest:
            self.implicit_subscription_rules.append((notification_name, rule_func))
        else:
            self.digest_implicit_subscription_rules.append((notification_name, rule_func))

    def add_alter_recipient_rule(self, notification_name, rule_func, digest=False):
        if not digest:
            self.alter_recipient_rules.append((notification_name, rule_func))
        else:
            self.digest_alter_recipient_rules.append((notification_name, rule_func))

    #--------------------------------------------------------------------------
    def subscribe(self, user, notification_name, sender=None, digest=False, force=False):
        """
        Add an observer to the subscription table
        :return:
        """
        subscription_info = {
            'user': user,
            'notification_name': notification_name,
            'digest': digest,
            }

        if isinstance(sender, Model):
            subscription_info['sender_type'] = ContentType.objects.get_for_model(sender)
            subscription_info['sender_id'] = sender.id

        subscription, created = Subscription.objects.get_or_create(**subscription_info)

        if created or force:
            subscription.is_active = True

        subscription.save()

    #--------------------------------------------------------------------------
    def unsubscribe(self, user, notification_name, sender=None, digest=False):
        """
        Remove an observer from the subscription table
        :return:
        """
        subscription_info = {
            'user': user,
            'notification_name': notification_name,
            'digest': digest,
        }

        if isinstance(sender, Model):
            subscription_info['sender_type'] = ContentType.objects.get_for_model(sender)
            subscription_info['sender_id'] = sender.id

        try:
            subscription = Subscription.objects.get(**subscription_info)
            subscription.is_active = False
            subscription.save()
        except Subscription.DoesNotExist:
            subscription_info['is_active'] = False
            Subscription.objects.create(**subscription_info)

    #--------------------------------------------------------------------------
    def subscriptions(self, notification_name, sender=None, digest=False):
        sender_filter = Q(sender_type__isnull=True)
        if sender:
            sender_filter |= Q(
                sender_type=ContentType.objects.get_for_model(sender),
                sender_id=sender.id,
            )

        digest_filter = Q(digest=digest)

        notification_name_filter = Q(notification_name=notification_name)
        notification_subname = notification_name

        while True:
            separator_index = notification_subname.rfind(':')
            if separator_index == -1:
                break
            notification_subname = notification_subname[:separator_index]
            notification_name_filter |= Q(notification_name=notification_subname)

        return Subscription.objects.filter(sender_filter & digest_filter & notification_name_filter)

    #--------------------------------------------------------------------------
    def active_subscriptions(self, notification_name, sender=None, digest=False):
        return self.subscriptions(notification_name, sender=sender, digest=digest).filter(is_active=True)

    #--------------------------------------------------------------------------
    def inactive_subscriptions(self, notification_name, sender=None, digest=False):
        return self.subscriptions(notification_name, sender=sender, digest=digest).filter(is_active=False)

    #--------------------------------------------------------------------------
    def subscribers(self, notification_name, sender=None, digest=False):
        """
        Return the list of users subscribed to a notification of a given name
        and optionally, from a given sender

        :param notification_name:
        :param sender:
        :return:
        """

        # Update recipient list with any applicable alteration rules
        subscribers = set()
        if not digest:
            implicit_subscription_rules = self.implicit_subscription_rules
        else:
            implicit_subscription_rules = self.digest_implicit_subscription_rules

        for implicit_subscription_notification_name, alter_subscribers_func in implicit_subscription_rules:
            if notification_name.startswith(implicit_subscription_notification_name):
                alter_subscribers_func(subscribers, notification_name, sender)

        # Get the list of explicit subscribers to the given notification
        active_subscriptions = self.active_subscriptions(notification_name, sender=sender, digest=digest)
        subscribers.update([s.user for s in active_subscriptions])

        # Exclude those who have explicitily unsubscribed
        inactive_subscriptions = self.inactive_subscriptions(notification_name, sender=sender, digest=digest)
        subscribers.difference_update([s.user for s in inactive_subscriptions])

        return subscribers

    def notification_classes(self, notification_name, sender=None, digest=False):
        try:
            notification_classes = getattr(sender, 'notification_classes')
        except AttributeError:
            return {}

        notification_type = notification_name.split(':')[-1]
        return notification_classes.get(notification_type , {})

    #--------------------------------------------------------------------------
    def post_notification(self, notification_name, created_by, sender=None, context=None):
        """
        :param notification_name:
        :param sender:
        :param context:
        :return:
        """
        context = context or {}
        try:
            date_scheduled = context.pop('date_scheduled')
        except KeyError:
            date_scheduled = None

        notification = Notification.objects.create(
            name=notification_name,
            created_by=created_by,
            sender=sender,
            context=context,
        )

        if os.environ.get('TESTING'):
            from .tasks import send_notification
            send_notification(notification.pk)
        else:
            if date_scheduled:
                schedule(
                    'notificationcenter.tasks.send_notification',
                    notification.pk,
                    schedule_type=Schedule.ONCE,
                    next_run=date_scheduled,
                )
            else:
                async_task(
                    'notificationcenter.tasks.send_notification',
                    notification.pk,
                )

    #--------------------------------------------------------------------------
    def notify(self, filter_params=None):
        """
        :return:
        """
        with PGAdvisoryLock("NotificationCenter.notify"):
            qs_params = {'date_sent__isnull': True}
            if filter_params:
                qs_params.update(filter_params)
            notifications = Notification.objects.filter(**qs_params).order_by('date_created')

            for notification in notifications:
                notification_context = notification.context

                # Determine the list of subscriber email addresses
                subscribers = self.subscribers(notification.name, sender=notification.sender)
                email_addresses = set([self.email_accessor(user) for user in subscribers])

                # Update email recipient list with any applicable alteration rules
                for notification_name, alter_recipients_func in self.alter_recipient_rules:
                    if notification.name.startswith(notification_name):
                        alter_recipients_func(email_addresses, notification)

                # Remove any falsy email address values that might've accumulated
                email_addresses.discard(None)
                email_addresses.discard('')

                # If there are any subscribers to this notification, send an email
                if email_addresses:
                    headers = notification_context.get('headers')
                    email = EmailMultiAlternatives(
                        subject=notification.subject,
                        body=notification.text,
                        from_email=self.from_email,
                        to=None,
                        bcc=email_addresses,
                        headers=headers
                    )

                    if notification.html:
                        email.attach_alternative(notification.html, "text/html")

                    try:
                        email.send(fail_silently=False)
                        notification.mark_as_sent({
                            'bcc': list(email_addresses),
                        })
                    except Exception as e:
                        # TODO: Log that the send failed
                        print(e)
                        pass
                else:
                    # If there are no subscribers to this notification, don't bother
                    # constructing an email. Simply mark it as sent.
                    notification.mark_as_sent()


    #---------------------------------------------------------------------------
    def digest(self, subject, grouping, template, recipients=None):
        """
        Send out a digest of all notifications for users that want a summary.

        :param grouping: a dictionary of notification name lists.

        digest({'active': ['purchasing:purchase_order:approve'], 'activity': ['purchasing:purchase_order:approve']})

        :return:
        """
        notifications = Notification.objects.filter(date_digested__isnull=True).order_by('date_created')
        digests = {}

        # When grouping, limit the notifications to the grouping set.
        names = []
        for key, ns in grouping:
            names = names + ns
        notifications = notifications.filter(
                name__in=names
            ).order_by(
                'sender_type', 'sender_id', 'name'
            )
        notifications = sorted(notifications, key=lambda x: x.date_created)

        for notification in notifications:
            subscribers = self.subscribers(notification.name, sender=notification.sender, digest=True)
            email_addresses = set([self.email_accessor(user) for user in subscribers])

            # Update email recipient list with any applicable alteration rules.
            for notification_name, alter_recipients_func in self.digest_alter_recipient_rules:
                if notification.name.startswith(notification_name):
                    alter_recipients_func(email_addresses, notification)

            email_addresses.discard(None)

            for email in email_addresses:
                digest = digests.get(email, collections.OrderedDict())
                for key, names in grouping:
                    group = digest.get(key, [])
                    if notification.name in names:
                        group.append(notification)
                    digest[key] = group
                digests[email] = digest

            # Mark this notification as having been digested.
            notification.date_digested = timezone.now()
            notification.save()

        # Send out the digested emails.
        for email, notifications in digests.items():

            # Scrub duplicate items.
            for key, values in notifications.items():
                unique = collections.OrderedDict()
                for notification in values:
                    unique[notification.sender] = notification
                notifications[key] = unique.values()

            context = {
                'today': datetime.date.today(),
                'notifications': notifications,
            }
            html = render_to_string(template, context)

            if (recipients and email in recipients) or recipients is None:
                email = EmailMultiAlternatives(
                    subject=subject,
                    body=strip_tags(html),
                    from_email=self.from_email,
                    to=[email],
                )
                email.attach_alternative(html, 'text/html')

                try:
                    email.send()
                except Exception:
                    pass

# A ready-to-import-and-use instance of the NotificationCenter
default_center = NotificationCenter()
if hasattr(settings, 'DEFAULT_NOTIFICATION_CENTER_EMAIL_ADDRESS_ACCESSOR'):
    default_center.email_accessor = import_string(settings.DEFAULT_NOTIFICATION_CENTER_EMAIL_ADDRESS_ACCESSOR)
if hasattr(settings, 'DEFAULT_NOTIFICATION_CENTER_PHONE_NUMBER_ACCESSOR'):
    default_center.phone_accessor = import_string(settings.DEFAULT_NOTIFICATION_CENTER_PHONE_NUMBER_ACCESSOR)



import traceback

from django.conf import settings
from django.core import mail
from django.core.mail import mail_admins
from django.db import connection
from django.db.utils import OperationalError

#-------------------------------------------------------------------------------
class PGAdvisoryLock(object):
    """
   Locks database operations on a sandbox with a pg_advisory lock.

   Use the same shared lock_value to ensure mutually exlusive access to
   the desired database records.
   """
    #---------------------------------------------------------------------------
    def __init__(self, lock_value):
        self.lock_hash = hash(lock_value)

    #---------------------------------------------------------------------------
    def __enter__(self):
        db_engine = settings.DATABASES['default']['ENGINE']
        if db_engine == 'django.db.backends.postgresql_psycopg2':
            cursor = connection.cursor()
            cursor.execute('SELECT pg_advisory_lock(%s)' % self.lock_hash)

    #---------------------------------------------------------------------------
    def __exit__(self, type, value, traceback):
        db_engine = settings.DATABASES['default']['ENGINE']
        if db_engine == 'django.db.backends.postgresql_psycopg2':
            cursor = connection.cursor()
            cursor.execute('SELECT pg_advisory_unlock(%s)' % self.lock_hash)


#------------------------------------------------------------------------------
class EmailCheck:
    """
    A context manager for use in email notification testing. Appearing within
    a test method as

        with EmailContext(self, users=[user1, ...]) as _:
            <block>

    or

        with EmailContext(self, contacts=[contact1, ...]) as _:
            <block>

    it will ensure that execution of <block> results in sending email to
    the given list of users/contacts. Note that exactly one of users and
    contacts must be provided. Note also that either users or contacts
    can be given as a function of one null argument; this is useful when
    the recipient list must be calculated after execution of the block.

    """
    def __init__(self, context, options, strict=True):
        self.context = context
        self.strict = strict
        self.options = options

        from .models import default_center as notification_center
        self.notification_center = notification_center

    def __enter__(self):
        # Clear the mail outbox
        mail.outbox = []
        return None

    def __exit__(self, exc_type, exc_value, trace):
        # Defer handling of any raised exceptions
        if exc_type:
            return False

        # Trigger notification delivery (is this really necessary?)
        self.notification_center.notify()

        # Ensure the mail outbox contains the expected message.
        messages = [msg for msg in mail.outbox]
        message_count = len(messages)

        recipients = self.options.get('recipients')

        if recipients:
            self.context.assertEqual(message_count, 1, "Unable to find expected message")
            message = messages[0]

            # Ensure that message has the expected bcc recipients
            expected = set()
            for recipient in recipients:
                if isinstance(recipient, str):
                    expected.add(recipient)
                else:
                    expected.add(self.notification_center.email_accessor(recipient))
            observed = set(message.recipients())
            self.context.assertEqual(expected - observed, set(), "missing recipients: %s" % (expected - observed))

            # Optionally require an exact match of bcc recipients
            if self.strict:
                self.context.assertEqual(observed - expected, set(), "extraneous recipients: %s" % (observed - expected))

            expected_subject = self.options.get('subject')
            if expected_subject:
                self.context.assertEqual(message.subject, expected_subject)

            expected_body_content = self.options.get('body')
            if expected_body_content:
                if not isinstance(expected_body_content, list):
                    expected_body_content = list(expected_body_content)

                for substring in expected_body_content:
                    self.context.assertIn(substring, message.body)

        else:
            self.context.assertEqual(message_count, 0, "An email was unexpectedly sent")

        return True


#-------------------------------------------------------------------------------
class NotifyOnError():
    def __enter__(self):
        pass
    def __exit__(self, exc_type, exc_value, exc_traceback):
        # If an error was raised, email it to the admins.
        # Return False, as we wish to allow the exception to be re-raised
        if (exc_type and exc_value and exc_traceback):
            message_body = '\n'.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
            mail_admins(
                "An uncaught exception has occurred",
                message_body,
                fail_silently=True
            )
        return False



from django.core.management.base import BaseCommand

#-------------------------------------------------------------------------------
class Command(BaseCommand):
    """
    Sends out any notifications in the NotificationCenter queue
    """
    def handle(self, *args, **options):
        from ...models import default_center
        default_center.notify()



# Generated by Django 3.1.3 on 2021-01-11 20:38

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion

class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender_id', models.PositiveIntegerField(null=True)),
                ('context', models.JSONField(default=dict)),
                ('name', models.CharField(max_length=100)),
                ('subject', models.CharField(max_length=255)),
                ('text', models.TextField(blank=True, null=True)),
                ('html', models.TextField(blank=True, null=True)),
                ('link', models.CharField(blank=True, max_length=255, null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('date_sent', models.DateTimeField(blank=True, null=True)),
                ('date_digested', models.DateTimeField(blank=True, null=True)),
                ('recipients', models.JSONField(blank=True, null=True)),
                ('sender_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='notification_sender_types', to='contenttypes.contenttype')),
            ],
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender_id', models.PositiveIntegerField(null=True)),
                ('notification_name', models.CharField(max_length=100)),
                ('is_active', models.BooleanField(default=True)),
                ('digest', models.BooleanField(default=False)),
                ('sender_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='notification_subscription_sender_types', to='contenttypes.contenttype')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('user', 'notification_name', 'sender_type', 'sender_id', 'digest')},
            },
        ),
    ]



import logging


from .models import (
    Notification,
    NotificationTypeChoices,
    default_center as notification_center,
)

def send_notification(notification_id):
    try:
        notification = Notification.objects.get(pk=notification_id)
    except Notification.DoesNotExist:
        # log(f'Notification not found (ID: {notification_id}', level=logging.ERROR)
        return True

    subscribers = notification_center.subscribers(notification.name, sender=notification.sender)
    if not subscribers:
        # If no users are subscribed to received this notification, there is nothing to do.
        return True

    notification_classes = notification_center.notification_classes(notification.name, sender=notification.sender)
    email_class = notification_classes.get(NotificationTypeChoices.EMAIL)
    if email_class:
        bcc = set([notification_center.email_accessor(user) for user in subscribers])
        bcc.discard(None)
        email = email_class(notification, bcc=bcc)
        email.send()

    sms_class = notification_classes.get(NotificationTypeChoices.SMS)
    if sms_class:
        recipients = set([notification_center.phone_accessor(user) for user in subscribers])
        recipients.discard(None)
        sms = sms_class(notification, recipients=recipients)
        sms.send()

    return True




from django.core import mail
from django.contrib.auth import get_user_model
from django.test import TestCase

from notificationcenter.models import NotificationCenter

User = get_user_model()
PASSWORD = '12345'

class NotificationCenterTests(TestCase):
    def setUp(self):
        # Create a few users to serve as observers to notification
        self.ricky = self._create_user('ricky', 'ricky@swearnet.com')
        self.julian = self._create_user('julian', 'julian@swearnet.com')
        self.bubbles = self._create_user('bubbles', 'bubbles@swearnet.com')

        self.notification_center = NotificationCenter()

    def _create_user(self, username, email):
        user, _ = User.objects.get_or_create(username=username, email=email)
        user.is_active = True
        user.set_password(PASSWORD)
        user.save()
        return user

    def test_event_notification(self):
        notification_name = 'trailerparkboys:cyrus_is_out_of_jail'
        notification_info = {
            'sender': None,
            'subject': "Cyrus was sprung from jail and he's coming for the driveway",
            'text': "Cyrus is on his way. We're gonna have a gunfight on our hands",
            'html': "<html><body>Cyrus is on his way. We're gonna have a gunfight on our hands</body></html>",
        }


        # Post a notification with no subscribers. Ensure no email is sent.
        mail.outbox = []
        self.notification_center.post_notification(notification_name, **notification_info)
        self.notification_center.notify()
        self.assertEqual(0, len(mail.outbox))

        # Subscribe a few users to that notification, post another notification, then process
        mail.outbox = []
        self.notification_center.subscribe(self.ricky, notification_name)
        self.notification_center.subscribe(self.julian, notification_name)
        self.notification_center.post_notification(notification_name, **notification_info)
        self.notification_center.notify()

        # Test that an email was sent to ricky and julian
        self.assertEqual(1, len(mail.outbox))
        message = mail.outbox[0]

        mail.outbox = []
        self.notification_center.unsubscribe(self.ricky, notification_name)
        self.notification_center.unsubscribe(self.julian, notification_name)
        self.notification_center.post_notification(notification_name, **notification_info)
        self.notification_center.notify()

        self.assertEqual(0, len(mail.outbox))

    def test_event_notification_with_sender(self):
        notification_name = 'trailerparkboys:hungry_kitty'

        purrmonster = self._create_user(username='purrmonster', email="purrmonster@kittylandlovecenter.com")
        shitrock = self._create_user(username='shitrock', email="purrmonster@kittylandlovecenter.com")

        # Bubbles wants to receive notifications about all hungry kitties
        self.notification_center.subscribe(self.bubbles, notification_name)

        # Ricky and Julian are only interested in notifications about specific hungry kitties
        self.notification_center.subscribe(self.ricky, notification_name, sender=purrmonster)
        self.notification_center.subscribe(self.julian, notification_name, sender=shitrock)

        purrmonster_notification_subject = "Ever since Lahey locked up food mountain, Purrmonster has been scavenging for scraps."
        self.notification_center.post_notification(
            notification_name,
            subject=purrmonster_notification_subject,
            sender=purrmonster
        )
        shitrock_notification_subject = "Shitrock has gone missing again and probably hasn't eaten in days."
        self.notification_center.post_notification(
            notification_name,
            subject=shitrock_notification_subject,
            sender=shitrock
        )
        self.notification_center.notify()

        # Ensure two notification emails were sent
        self.assertEqual(2, len(mail.outbox))

        # Verify that only bubbles and ricky received a notification about purrmonster
        purrmonster_message = mail.outbox[0]
        self.assertEqual(purrmonster_message.subject, purrmonster_notification_subject)
        self.assertEqual(set(purrmonster_message.recipients()), set([self.ricky.email, self.bubbles.email]))

        # Verify that only Bubbles and Julian received a notification about shitrock
        shitrock_message = mail.outbox[1]
        self.assertEqual(shitrock_message.subject, shitrock_notification_subject)
        self.assertEqual(set(shitrock_message.recipients()), set([self.julian.email, self.bubbles.email]))

    def test_event_notification_digest(self):
        mail.outbox = []
        notification_name = 'trailerparkboys:cyrus_is_out_of_jail'

        # Subscribe a few digest users to the notifications.
        self.notification_center.subscribe(self.ricky, notification_name, digest=True)
        self.notification_center.subscribe(self.julian, notification_name, digest=True)

        notification_info = {
            'sender': None,
            'subject': "Cyrus was sprung from jail and he's coming for the driveway",
            'text': "Cyrus is on his way. We're gonna have a gunfight on our hands",
            'html': "<html><body>Cyrus is on his way. We're gonna have a gunfight on our hands</body></html>",
            'link': "https://www.wikipedia.org"
        }
        self.notification_center.post_notification(notification_name, **notification_info)
        self.notification_center.notify()
        self.assertEqual(len(mail.outbox), 0)

        notification_info = {
            'sender': None,
            'subject': "Just a false alarm, Trever is an idiot",
            'text': "Trevor is an idiot, he was confused about seeing Cyrus",
            'html': "<html><body>Trevor is an idiot, he was confused about seeing Cyrus</body></html>",
            'link': "https://www.wikipedia.org"
        }
        self.notification_center.post_notification(notification_name, **notification_info)
        self.notification_center.notify()
        self.assertEqual(len(mail.outbox), 0)

        self.notification_center.digest(
            'Test Digest', [
                ('test', ['trailerparkboys:cyrus_is_out_of_jail'])
            ], 'notificationcenter/digest.html'
        )
        self.assertEqual(len(mail.outbox), 2)

